//FIXME:reading from stdin;
//FIXME:setlocale;

#include<limits.h>
#include<stdlib.h>
#include<assert.h>
#include<iostream>
#include<string>
#include<sstream>
#include<list>
#include<vector>
#include<set>
#include"musitoriusclient.h"
#include"CmdArgsParser.h"
#include<string-utils.h>
#include<errno.h>
#include<string.h>
#include<sys/types.h>
#include<unistd.h>
#include<fcntl.h>
#include<sys/stat.h>

#define ERRNO_MSG (strerror(errno))
#define PREFIX "musitorius:"
#define PLAYLIST_EXTENSION "m3u"

typedef std::list<std::string> StringList;

static CmdArg cmdLineParams[] = {
  {'b', "backward", "STEP", "jump backward onto STEP seconds;"},
  {'f', "forward", "STEP", "jump backward onto STEP seconds;"},
  {'h', "help", "", "print this help screen and exit;"},
  {'n', "next", "", "go to the next playlist item;"},
  {'l', "listen", "", "listen and print server state changes;"},
  {'p', "play", "", "toggle play/pause state;"},
  {'L', "playlist", "", "process all input files as playlists;"},
  {'P', "prev", "", "go to the previous playlist item;"},
  {'s', "stop", "", "stop the current playback;"},
  {' ', NULL, NULL, NULL}
};

CmdArgsParser cmdLine(cmdLineParams);

std::string readTextFile(const std::string& fileName)
{
  const int fd = open(fileName.c_str(), O_RDONLY);
  if (fd == -1)
    {
      std::cerr << PREFIX << "open(" << fileName << "):" << ERRNO_MSG << std::endl;
      exit(EXIT_FAILURE);
    }
  std::string s;
  char buf[2048];
  ssize_t readCount;
  do {
    readCount = read(fd, buf, sizeof(buf));
    if (fd == -1)
      {
	std::cerr << PREFIX << "read(" << fileName << "):" << ERRNO_MSG << std::endl;
	exit(EXIT_FAILURE);
      }
    for(ssize_t i = 0;i < readCount;i++)
      s += buf[i];
  } while(readCount);
  close(fd);
  return s;
}

void readPlaylist(const std::string& fileName, StringList& items)
{
  const std::string text = readTextFile(fileName);
  std::string s;
  for(std::string::size_type i = 0;i < text.length();i++)
    {
      const char c = text[i];
      if (c == '\r')
	continue;
      if (c == '\n')
	{
	  s = trim(s);
	  if (s.empty())
	    continue;
	  if (s[0] == '#')
	    {
	      s.erase();
	      continue;
	    }
	  items.push_back(s);
	  s.erase();
	  continue;
	}
      s += c;
    }
}

std::string fileNameExtension(const std::string& fileName)
{
  std::string s;
  for(std::string::size_type i = 0;i < fileName.length();i++)
    {
      if (fileName[i] == '.')
	{
	  s.erase();
	  continue;
	}
      s += fileName[i];
    }
  return s;
}

std::string ensureAbsolutePath(const std::string& fileName)
{
  char resolved[PATH_MAX];
  if (realpath(fileName.c_str(), resolved))
    return resolved;
  std::cerr << PREFIX << fileName << ":" << ERRNO_MSG << std::endl;
  return "";
}

void printError(musitorius_result_t res)
{
  if (res == -1)
    std::cerr << PREFIX << "command failed:" << ERRNO_MSG << std::endl; else
    std::cerr << PREFIX << "command failed:see server logs for detailed information" << std::endl;
}

void printHelp()
{
  std::cout << "musitorius: console client for Musitorius multimedia player control utility; Version: " << PACKAGE_VERSION << std::endl << std::endl;
  std::cout << "Usage:" << std::endl << "\tmusitorius [OPTIONS] [FILE1 [FILE2 [...]]]" <<std::endl << std::endl;
  cmdLine.printHelp();
}

int main(int argc, char* argv[])
{
  //FIXME:automatic connection closing;
  if (!cmdLine.parse(argc, argv))
    return 1;
  if (cmdLine.used("help"))
    {
      printHelp();
      return 0;
    }
  musitorius_connection_t connection = musitorius_connect((char*)MUSITORIUS_DEFAULT_SOCKET);//FIXME:
  if (connection == MUSITORIUS_BAD_CONNECTION)
    {
      std::cerr << PREFIX << "connection could not be established" << std::endl;
      return 1;
    }

  if (cmdLine.used("play"))
    {
      musitorius_result_t res = musitorius_play(connection);
      if (!MUSITORIUS_OK(res))
	{
	  printError(res);
	  return 1;
	}
      return 0;
    }

  if (cmdLine.used("stop"))
    {
      musitorius_result_t res = musitorius_stop(connection);
      if (!MUSITORIUS_OK(res))
	{
	  printError(res);
	  return 1;
	}
      return 0;
    }

  if (cmdLine.used("next"))
    {
      musitorius_result_t res = musitorius_next(connection);
      if (!MUSITORIUS_OK(res))
	{
	  printError(res);
	  return 1;
	}
      return 0;
    }

  if (cmdLine.used("prev"))
    {
      musitorius_result_t res = musitorius_prev(connection);
      if (!MUSITORIUS_OK(res))
	{
	  printError(res);
	  return 1;
	}
      return 0;
    }

  if (cmdLine.used("backward"))
    {
      const std::string value = trim(cmdLine["backward"]);
      if (!checkTypeUnsignedInt(value))
	{
	  std::cerr << PREFIX << "\'" << value << "\' is not a valid seconds number" << std::endl;
	    return 1;
	}
      const size_t step = parseAsUnsignedInt(value);
      musitorius_result_t res = musitorius_backward(connection, step);
      if (!MUSITORIUS_OK(res))
	{
	  printError(res);
	  return 1;
	}
      return 0;
    }

  if (cmdLine.used("forward"))
    {
      const std::string value = trim(cmdLine["forward"]);
      if (!checkTypeUnsignedInt(value))
	{
	  std::cerr << PREFIX << "\'" << value << "\' is not a valid seconds number" << std::endl;
	    return 1;
	}
      const size_t step = parseAsUnsignedInt(value);
      musitorius_result_t res = musitorius_forward(connection, step);
      if (!MUSITORIUS_OK(res))
	{
	  printError(res);
	  return 1;
	}
      return 0;
    }

  if (cmdLine.files.empty())
    {
      printHelp();
      return 0;
    }
  const bool everyAsPlaylist = cmdLine.used("playlist");
  StringList items;
  for(size_t i = 0;i < cmdLine.files.size();i++)
    {
      const std::string& value = cmdLine.files[i];
      if (trim(value).empty())
	continue;
      if (everyAsPlaylist)
	{
	  readPlaylist(value, items);
	  continue;
	}
      if (value.length() > 3 && fileNameExtension(value) == PLAYLIST_EXTENSION)
	{
	  readPlaylist(value, items);
	  continue;
	}
      items.push_back(value);
    }
  musitorius_string_list_t stringList;
  musitorius_string_list_init(&stringList);
  for(StringList::const_iterator it = items.begin();it != items.end();++it)
    {
      const std::string absName = ensureAbsolutePath(*it);
      if (!absName.empty())
	musitorius_string_list_add(&stringList, (char*)absName.c_str());
    }
  musitorius_result_t res = musitorius_task(connection, &stringList);
  musitorius_string_list_destroy(&stringList);
  if (!MUSITORIUS_OK(res))
    {
      printError(res);
      return 1;
    }
  return 0;
}
