
#ifndef __MUSITORIUS_CLIENT_H__
#define __MUSITORIUS_CLIENT_H__

#include"Task.h"
#include"Protocol.h"

class ClientCallback//FIXME:Abstract;
{
public:
  //Query methods;
  virtual bool onQueryIsIdle() const = 0;
  virtual bool onQueryCurrentTask(Task& task) const = 0;
  virtual TaskItemId onQueryCurrentTaskItemId() const = 0;
  virtual bool onQueryCurrentPos(size_t& pos, size_t& total) const = 0;
  virtual size_t onQueryVolume() const = 0;

  //Command methods;
  virtual void onNewTask(const Task& task) = 0;
  virtual bool onPlayToggle() = 0;
  virtual void onStop() = 0;
  virtual void onForward(size_t offset, size_t& newPos, size_t& total) = 0;
  virtual void onBackward(size_t offset, size_t& newPos, size_t& total) = 0;
  virtual TaskItemId onNext() = 0;
  virtual TaskItemId onPrevious() = 0;
  virtual bool onTaskItem(TaskItemId taskItemId) = 0;
  virtual size_t onIncreaseVolume(size_t step) = 0;
  virtual size_t onDecreaseVolume(size_t step) = 0;
}; //class CLientCallback;

class Client: private AbstractProtocolHandler, private AbstractProtocolOutput
{
public:
  Client(ClientCallback& callback, int fd) 
    : m_callback(callback), m_fd(fd), m_listenMode(0), m_protocol(*this, *this) {}

  virtual ~Client();

  bool readData();
  void close();
  int getHandler() const;

  void notifyOnNewTask(const Task& task);
  void notifyOnNewTaskItem(TaskItemId taskItemId);
  void notifyOnNewPos(size_t pos, size_t total);
  void notifyOnIdle();
  void notifyOnTaskFinish();
  void notifyOnNewVolume(size_t value);

private: //AbstractProtocolHandler;
  void onCommandQueryIsIdle(ServerReply& reply);
  void onCommandQueryCurrentTask(ServerReply& reply);
  void onCommandQueryCurrentTaskItemId(ServerReply& reply);
  void onCommandQueryCurrentPos(ServerReply& reply);
  void onCommandNewTask(const Task& task, ServerReply& reply);
  void onCommandPlay(ServerReply& reply);
  void onCommandStop(ServerReply& reply);
  void onCommandForward(size_t offset, ServerReply& reply);
  void onCommandBackward(size_t offset, ServerReply& reply);
  void onCommandNext(ServerReply& reply);
  void onCommandPrevious(ServerReply& reply);
  void onCommandTaskItem(TaskItemId taskItemId, ServerReply& reply);
  void onCommandIncreaseVolume(size_t step, ServerReply& reply);
  void onCommandDecreaseVolume(size_t step, ServerReply& reply);
  void onCommandSwitchToListen();

private: //AbstractProtocolOutput;
  void onProtocolOutput(const std::string& output);

private:
  ClientCallback& m_callback;
  int m_fd;
  bool m_listenMode;
  Protocol m_protocol;
  std::string m_inputChain;
}; //class Client;

typedef std::list<Client*> ClientList;
typedef std::vector<Client*> ClientVector;

#endif //__MUSITORIUS_CLIENT_H__;
