
#include"musitorius.h"
#include"MainLoop.h"

void MainLoop::run(const SocketList& sockets, sigset_t* sigMask)
{
  assert(sigMask != NULL);
  logMsg(LOG_DEBUG, "Main server loop started!");
  while(!m_terminationFlag)
    {
      fd_set fds;
      SocketList::const_iterator socketIt;
      ClientList::iterator clientIt;
      int maxFd = 0;
      FD_ZERO(&fds);
      for(socketIt = sockets.begin();socketIt != sockets.end();socketIt++)
	{
	  const int fd=(*socketIt)->getHandler();
	  if (fd > maxFd)
	    maxFd = fd;
	  FD_SET(fd, &fds);
	} //for (sockets);
      for(clientIt = m_connectedClients.begin();clientIt != m_connectedClients.end();clientIt++)
	{
	  const int fd=(*clientIt)->getHandler();
	  if (fd > maxFd)
	    maxFd = fd;
	  FD_SET(fd, &fds);
	} //for (clients);
      const int playerStdout = m_playerOutput.getOutputFd();
      const int playerStderr = m_playerOutput.getErrorFd();
      FD_SET(playerStdout, &fds);
      FD_SET(playerStderr, &fds);
      if (maxFd < playerStdout)
	maxFd = playerStdout;
      if (maxFd < playerStderr)
	maxFd = playerStderr;
      if (pselect(maxFd + 1, &fds, NULL, NULL, NULL, sigMask) == -1)
	{
	  const int errorCode = errno;
	  logMsg(LOG_DEBUG, "pselect() has returned -1, checking what the reason");
	  if (errorCode == EINTR)
	    {
	      logMsg(LOG_DEBUG, "pselect() call was interrupted by system signal, going to next iteration");
	      m_signalHandler.onSystemSignal();
	      continue;
	    } //EINTR;
	  logMsg(LOG_DEBUG, "pselect() has returned an unexpected error, stopping main loop ");
	  throw SystemException(errorCode, "pselect()");
	} //select has returned an error;
      bool wasNewClients = 0;
      for(socketIt = sockets.begin();socketIt != sockets.end();socketIt++)
	{
	  const int fd = (*socketIt)->getHandler();
	  if (!FD_ISSET(fd, &fds))
	    continue;
	  //OK:we found new connected client;
	  logMsg(LOG_DEBUG, "New connection registered");
	  wasNewClients=1;
	  const int newClientFd = ::accept(fd, NULL, NULL);
	  if (newClientFd == -1)
	    {
	      logMsg(LOG_ERR, "New client cannot be accepted, accept() says \'%s\'", ERRNO_MSG);
	      continue;
	    }
	  logMsg(LOG_DEBUG, "New client connection was established (fd=%d)", newClientFd);
	  if (m_maxClients > 0 && m_connectedClients.size() >= m_maxClients)
	    {
	      ::close(newClientFd);
	      logMsg(LOG_WARNING, "Client count limit reached, closing new connection (already have %u clients)", m_connectedClients.size());
	    } else//adding new record to the connected clients list;
	    {
	      auto_ptr<Client> newClient = m_clientFactory.createNewClient(newClientFd);
	      m_connectedClients.push_back(newClient.get());
	      newClient.release();
	      logMsg(LOG_INFO, "New connection was successfully accepted and added to the list of connected clients (fd=%d)", newClientFd);
	    }
	} //new clients handling;
      if (wasNewClients)
	continue;//resetting set of file descriptors with new sockets;
      //now checking list of clients, is there any data from them;
      for(clientIt = m_connectedClients.begin();clientIt != m_connectedClients.end();clientIt++)
	{
	  Client* client = *clientIt;
	  assert(client);
	  const int fd = client->getHandler();
	  if (!FD_ISSET(fd, &fds))
	    continue;
	  const bool toClose = client->readData();
	  if (toClose)
	    {
	      client->close();
	      delete client;
	      m_connectedClients.erase(clientIt);
	      logMsg(LOG_INFO, "Client was closed and its data destroyed (fd=%d)", fd);
	      break;//do not handle other clients;
	    } // closing connection;
	} // for (clients);
      if (FD_ISSET(playerStdout, &fds))
	m_playerOutput.readOutputData();
      if (FD_ISSET(playerStderr, &fds))
	m_playerOutput.readErrorData();
    } // while(!m_terminationFlag);
  for(ClientList::iterator it = m_connectedClients.begin();it != m_connectedClients.end();it++)
    delete *it;
  m_connectedClients.clear();
}
