
//FIXME:check error reply on errors;

#include"musitorius.h"
#include"Protocol.h"

/*
 * The list of unimplemented commands in protocol:
 *
 * void onCommandQueryIsIdle(ServerReply& reply);
 * void onCommandQueryCurrentTask(ServerReply& reply);
 * void onCommandQueryCurrentTaskItemId(ServerReply& reply);
 * void onCommandQueryCurrentPos(ServerReply& reply);
 * void onCommandTaskItem(TaskItemId taskItemId, ServerReply& reply);
 * void onCommandIncreaseVolume(size_t step, ServerReply& reply);
 * void onCommandDecreaseVolume(size_t step, ServerReply& reply);
 * void onCommandSwitchToListen();
 */

static std::string cutHead(const std::string& line)
{
  if (line.length() < 2)
    return "";
  return line.substr(1);
}

static void splitCommand(const std::string& src, std::string& command, std::string& params)
{
  command.erase();
  params.erase();
  if (src.empty())
    return;
  bool wasSpace = 0;
  for(std::string::size_type i = 0;i < src.length();i++)
    {
      const char c = src[i];
      if (wasSpace)
	{
	  params += c;
	  continue;
	}
      if (BLANK_CHAR(c))
	{
	  wasSpace = 1;
	  continue;
	}
      command += c;
    }
}

void Protocol::processNewTaskCommand(const StringList& arg)
{
  logMsg(LOG_DEBUG, "Processing new task command with %u items", arg.size());
  Task task;
  task.initAsFileNames(arg);
  ServerReply reply;
  m_handler.onCommandNewTask(task, reply);
  assert(reply.type == ServerReplyBoolean);
  sendGeneralBooleanReply(reply.booleanValue);
}

void Protocol::processPlayCommand(const StringList& arg)
{
  if (!arg.empty())
    logMsg(LOG_WARNING, "received play command, but argument list is not empty");
  logMsg(LOG_DEBUG, "Processing play command");
  ServerReply reply;
  m_handler.onCommandPlay(reply);
  assert(reply.type == ServerReplyBoolean);
  //Really we know here what is the new server status through reply object but do not provide this information to client in this version, just OK answer;
  sendGeneralBooleanReply(1);
}

void Protocol::processStopCommand(const StringList& arg)
{
  if (!arg.empty())
    logMsg(LOG_WARNING, "received stop command, but argument list is not empty");
  logMsg(LOG_DEBUG, "Processing stopping command");
  ServerReply reply;
  m_handler.onCommandStop(reply);
  assert(reply.type == ServerReplyBoolean);
  sendGeneralBooleanReply(reply.booleanValue);
}

void Protocol::processNextCommand(const StringList& arg)
{
  if (!arg.empty())
    logMsg(LOG_WARNING, "received next task item command, but argument list is not empty");
  logMsg(LOG_DEBUG, "Processing command to go to next task item");
  ServerReply reply;
  m_handler.onCommandNext(reply);
  assert(reply.type == ServerReplyBoolean);
  sendGeneralBooleanReply(reply.booleanValue);
}

void Protocol::processPrevCommand(const StringList& arg)
{
  if (!arg.empty())
    logMsg(LOG_WARNING, "received previous task item command, but argument list is not empty");
  logMsg(LOG_DEBUG, "Processing command to go to previous task item");
  ServerReply reply;
  m_handler.onCommandPrevious(reply);
  assert(reply.type == ServerReplyBoolean);
  sendGeneralBooleanReply(reply.booleanValue);
}

void Protocol::processBackwardCommand(const std::string& param, const StringList& arg)
{
  size_t step = 0;
  if (!checkTypeUnsignedInt(trim(param)))
    {
      logMsg(LOG_ERR, "Received from client invalid parameter for backward command: \'%s\'", param.c_str());
      sendGeneralBooleanReply(0);
      return;
    }
  step = parseAsUnsignedInt(param);
  if (!arg.empty())
    logMsg(LOG_WARNING, "received backward command, but argument list is not empty");
  logMsg(LOG_DEBUG, "Processing command to make backward step onto %u seconds", step);
  ServerReply reply;
  m_handler.onCommandBackward(step, reply);
  assert(reply.type == ServerReplyBoolean);
  sendGeneralBooleanReply(reply.booleanValue);
}

void Protocol::processForwardCommand(const std::string& param, const StringList& arg)
{
  size_t step = 0;
  if (!checkTypeUnsignedInt(trim(param)))
    {
      logMsg(LOG_ERR, "Received from client invalid parameter for forward command: \'%s\'", param.c_str());
      sendGeneralBooleanReply(0);
      return;
    }
  step = parseAsUnsignedInt(param);
  if (!arg.empty())
    logMsg(LOG_WARNING, "received forward command, but argument list is not empty");
  logMsg(LOG_DEBUG, "Processing command to make forward step onto %u seconds", step);
  ServerReply reply;
  m_handler.onCommandForward(step, reply);
  assert(reply.type == ServerReplyBoolean);
  sendGeneralBooleanReply(reply.booleanValue);
}

void Protocol::handleCommand()
{
  logMsg(LOG_DEBUG, "Processing command with %u lines", m_collectedCommand.size());
  assert(!m_collectedCommand.empty());
  const std::string firstLine = m_collectedCommand.front();
  m_collectedCommand.pop_front();
  std::string command, params;
  splitCommand(firstLine, command, params);
  command = toLower(trim(command));

  if (command == "task")
    processNewTaskCommand(m_collectedCommand);

  if (command == "play")
    processPlayCommand(m_collectedCommand);

  if (command == "stop")
    processStopCommand(m_collectedCommand);

  if (command == "next")
    processNextCommand(m_collectedCommand);

  if (command == "prev")
    processPrevCommand(m_collectedCommand);

  if (command == "backward")
    processBackwardCommand(params, m_collectedCommand);

  if (command == "forward")
    processForwardCommand(params, m_collectedCommand);

  m_collectedCommand.clear();
}

void Protocol::sendNotification(char notificationType, const ServerReply& reply)
{
  //FIXME:
}

void Protocol::processInputLine(const std::string& inputLine)
{
  logMsg(LOG_DEBUG, "Protocol handler is processing new client string \'%s\'", inputLine.c_str());
  //FIXME:command lines number limit;
  if (inputLine.empty())
    {
      logMsg(LOG_WARNING, "Protocol handler is requested to process an empty input line");
      return;
    }
  const char head = inputLine[0];
  if (head != '=' && head != '+' && head != '-')
    {
      logMsg(LOG_WARNING, "The protocol handler is requested to process line with invalid head \'%s\'", inputLine.c_str());
      return;
    }
  if (m_readyToNextCommand)
    {
      m_collectedCommand.clear();
      if (head != '=')
	{
	  logMsg(LOG_WARNING, "New command in protocol can be started only with \'=\' head, the following line is invalid: \'%s\'", inputLine.c_str());
	  return;
	}
      m_collectedCommand.push_back(cutHead(inputLine));
      m_readyToNextCommand = 0;
      return;
    } //new command;
  assert(!m_readyToNextCommand);
  if (head == '+')
    {
      //FIXME:lines number limit;
      m_collectedCommand.push_back(cutHead(inputLine));
      return;
    }
  assert(head == '-');
  handleCommand();
  m_collectedCommand.clear();
  m_readyToNextCommand = 1;
}

void Protocol::sendGeneralBooleanReply(char value)
{
  if (value)
    m_output.onProtocolOutput("200:OK!"); else
    m_output.onProtocolOutput("400:Failed. For further information see server logs.");

}
