

#ifndef __MUSITORIUS_PROTOCOL_H__
#define __MUSITORIUS_PROTOCOL_H__

#include"Task.h"

enum {
  ServerReplyBoolean = 0,
  ServerReplyPos = 1,
  ServerReplyPercent = 2
};

enum {
  NotificationNewTask = 0,
  NotificationNewTaskItem = 1,
  NotificationNewPos = 2,
  NotificationIdle = 3
};

class ServerReply
{
public:
  ServerReply()
    : type(ServerReplyBoolean), booleanValue(0), posValue1(0), posValue2(0) {}

  char type;
  bool booleanValue;
  size_t posValue1, posValue2;
  size_t percentValue;
}; //class ServerReply;

class AbstractProtocolHandler
{
public:
  virtual void onCommandQueryIsIdle(ServerReply& reply) = 0;
  virtual void onCommandQueryCurrentTask(ServerReply& reply) = 0;
  virtual void onCommandQueryCurrentTaskItemId(ServerReply& reply) = 0;
  virtual void onCommandQueryCurrentPos(ServerReply& reply) = 0;
  virtual void onCommandNewTask(const Task& task, ServerReply& reply) = 0;
  virtual void onCommandPlay(ServerReply& reply) = 0;
  virtual void onCommandStop(ServerReply& reply) = 0;
  virtual void onCommandForward(size_t offset, ServerReply& reply) = 0;
  virtual void onCommandBackward(size_t offset, ServerReply& reply) = 0;
  virtual void onCommandNext(ServerReply& reply) = 0;
  virtual void onCommandPrevious(ServerReply& reply) = 0;
  virtual void onCommandTaskItem(TaskItemId taskItemId, ServerReply& reply) = 0;
  virtual void onCommandIncreaseVolume(size_t step, ServerReply& reply) = 0;
  virtual void onCommandDecreaseVolume(size_t step, ServerReply& reply) = 0;
  virtual void onCommandSwitchToListen() = 0;
}; //class AbstractProtocolHandler;

class AbstractProtocolOutput
{
public:
  virtual void onProtocolOutput(const std::string& output) = 0;
}; //class AbstractProtocolOutput;

class Protocol
{
public:
  Protocol(AbstractProtocolHandler& handler, AbstractProtocolOutput& output)
    : m_handler(handler), m_output(output), m_readyToNextCommand(1) {}

  void sendNotification(char notificationType, const ServerReply& reply);
  void processInputLine(const std::string& inputLine);

private:
  void handleCommand();
  void processBackwardCommand(const std::string& param, const StringList& arg);
  void processForwardCommand(const std::string& param, const StringList& arg);
  void processNewTaskCommand(const StringList& arg);
  void processNextCommand(const StringList& arg);
  void processPlayCommand(const StringList& arg);
  void processPrevCommand(const StringList& arg);
  void processStopCommand(const StringList& arg);
  void sendGeneralBooleanReply(char value);

private:
  AbstractProtocolHandler& m_handler;
  AbstractProtocolOutput& m_output;
  StringList m_collectedCommand;
  bool m_readyToNextCommand;
}; //class Protocol;

#endif //__MUSITORIUS_PROTOCOL_H__;
