
#include"system.h"
#include"string-utils.h"

#define MUSITORIUS_LOGGER "musitorius"

static bool configLogConsole=1;
static int configLogLevel=LOG_WARNING;

static void printLogLine(int level, const char* line)
{
  if (!configLogConsole)
    return;
  if (level <= LOG_CRIT)
    {
      std::cerr << "FATAL:" << line << std::endl;
      return;
    }
  if (level <= LOG_ERR)
    {
      std::cerr << "ERROR:" << line << std::endl;
      return;
    }
  if (level <= LOG_WARNING)
    {
      std::cerr << "WARN:" << line << std::endl;
      return;
    }
  std::cout << line << std::endl;
}

static void saveLogLine(int level, const char* line)
{
  assert(line != NULL);
  syslog(level, "%s", line);
}

static void logLine(int level, const char* line)
{
  if (level > configLogLevel)
    return;
  printLogLine(level, line);
  saveLogLine(level, line);
}

void logMsg(int level, const char* format, ...)
{
  if (!format)
    return;
  va_list args;
  va_start(args, format);
  char buf[4096];
  vsnprintf(buf, sizeof(buf), format, args);
  buf[sizeof(buf)-1]='\0';
  va_end(args);
  std::string str(buf);
  removeNewLineChars(str);
  logLine(level, str.c_str());
}

void initLogging(bool logConsole, int logLevel)
{
  configLogConsole = logConsole;
  configLogLevel = logLevel;
  openlog(MUSITORIUS_LOGGER, LOG_PID, LOG_DAEMON);
}
