//FIXME:SIGPIPE;
//FIXME:setlocale;

#include"musitorius.h"
#include"core/core.h"
#include"CmdArgsParser.h"
#include"ConfigFile.h"

#define PREFIX "musitorius:"
#define MAX_VOLUME_VALUE 100

static CmdArg cmdLineParams[] = {
  {'c', "conf", "FILE_NAME", "read server configuration from FILE_NAME;"},
  {'d', "daemon", "", "run server as a daemon;"},
  {'h', "help", "", "print this help screen and exit;"},
  {'p', "pidfile", "FILE_NAME", "save pid of server process to FILE_NAME;"},
  {'s', "socket", "FILE_NAME", "open UNIX domain socket at FILE_NAME."},
  {' ', NULL, NULL, NULL}
};

CmdArgsParser cmdLine(cmdLineParams);

volatile sig_atomic_t wasSigChld = 0;
void sigChldHandler(int r)
{
  wasSigChld = 1;
}

enum {
  StatusFree = 0,
  StatusPlaying = 1,
StatusPaused = 2
};

class Configuration
{
public:
  Configuration() 
    : socket(MUSITORIUS_DEFAULT_SOCKET), daemonMode(0) {}

  std::string pidFile;
  std::string socket;
  bool daemonMode;
}; //class Configuration;

class ConsoleStatus
{
public:
  ConsoleStatus() {}
  virtual ~ConsoleStatus() {}

  void notifyOnNewTask(const Task& task)
  {
    //FIXME:
  }

  void notifyOnNewTaskItem(const Task& task, TaskItemId taskItemId)
  {
    //FIXME:
  }

  void notifyOnNewPos(size_t pos, size_t total)
  {
    //FIXME:
  }

  void notifyOnIdle()
  {
    //FIXME:
  }

  void notifyOnTaskFinish()
  {
    //FIXME:
  }

  void notifyOnNewVolume(size_t value)
  {
    //FIXME:
  }
}; //class ConsoleStatus;

class MusitoriusServer: private ClientCallback, private AbstractSignalHandler, private AbstractPlayerCallback
{
public:
  MusitoriusServer(const Configuration& c)
    : m_configuration(c), m_status(StatusFree), m_currentTaskItemId(BAD_TASK_ITEM_ID), m_currentVolume(MAX_VOLUME_VALUE / 2), m_terminationFlag(0)  {}
  virtual ~MusitoriusServer() 
  {
    closeSockets();
  }

  void run()
  {
    installSignalProcessing();
    auto_ptr<Socket> socket(new Socket());
    socket->open(m_configuration.socket);
    logMsg(LOG_DEBUG, "UNIX domain socket is opened as \'%s\'", socket->getFileName().c_str());
    m_sockets.push_back(socket.get());
    socket.release();
    m_player = createMPlayerPlayer("mplayer -slave", *this);//FIXME:
    m_player->init();
    ClientFactory clientFactory(*this);
    MainLoop mainLoop(clientFactory, m_clients, 10, *this, *m_player.get(), m_terminationFlag);//FIXME:10 - limited number of clients;
    logMsg(LOG_DEBUG, "Ready to run main loop!");
    mainLoop.run(m_sockets, &m_origMask);
  }

private: //AbstractClientCallback;
  bool onQueryIsIdle() const
  {
    logMsg(LOG_DEBUG, "The server is requested for idle status, answer is %d", m_status != StatusPlaying);
    return m_status != StatusPlaying;
  }

  bool onQueryCurrentTask(Task& task) const
  {
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested for current task, but there is no any");
	return 0;
      }
    logMsg(LOG_DEBUG, "The server is requested for current task");
    task = m_currentTask;
    return 1;
  }

  TaskItemId onQueryCurrentTaskItemId() const
  {
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested for current task ID, but there is no any task at all");
	return BAD_TASK_ITEM_ID;
      }
    logMsg(LOG_DEBUG, "The server is requested for current task ID, answer is %u", m_currentTaskItemId);
    return m_currentTaskItemId;
  }

  bool onQueryCurrentPos(size_t& pos, size_t& total) const
  {
    if (m_status == StatusPlaying)
      {
	logMsg(LOG_DEBUG, "The server is requested for current position and task item total length, but there is no any task");
	return 0;
      }
    logMsg(LOG_DEBUG, "The server is requested for current position and task item total length, answer is %u and %u", m_lastPos, m_totalSec);
    pos = m_lastPos;
    total = m_totalSec;
    return 1;
  }

  size_t onQueryVolume() const
  {
    logMsg(LOG_DEBUG, "The server is requested for current volume level, answer is %u", m_currentVolume);
    return m_currentVolume;
  }

  void onNewTask(const Task& task)
  {
    logMsg(LOG_DEBUG, "The server received command to initiate new task  playback");
    if (m_status != StatusFree)
      {
	logMsg(LOG_INFO, "Closing previously opened task to be ready for new one");
	m_player->stop();
	m_status = StatusFree;
	m_currentTaskItemId = BAD_TASK_ITEM_ID;
	m_lastPos = 0;
	m_totalSec = 0;
	for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
	  (*it)->notifyOnTaskFinish();
	m_consoleStatus.notifyOnTaskFinish();
      }
    m_currentTask.clear();
    if (task.empty())
      {
	logMsg(LOG_DEBUG, "New task is empty, doing nothing");
	return;
      }
    m_currentTask = task;
    m_currentTaskItemId = m_currentTask.startingItemId();
    if (m_currentTaskItemId == BAD_TASK_ITEM_ID)
      {
	logMsg(LOG_WARNING, "Newly created task cannot choose the starting item");
	return;
      }
    m_status = StatusPlaying;
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
      (*it)->notifyOnNewTask(m_currentTask);
    m_consoleStatus.notifyOnNewTask(m_currentTask);
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
      (*it)->notifyOnNewTaskItem(m_currentTaskItemId);
    m_consoleStatus.notifyOnNewTaskItem(m_currentTask, m_currentTaskItemId);
    logMsg(LOG_INFO, "Starting new task with %u items", m_currentTask.getItemCount());
    goPlay();
  }

bool onPlayToggle()
  {
    if (m_currentTask.empty())
      {
	if (m_status != StatusFree);
	logMsg(LOG_WARNING, "requested play toggle operation: task is empty but status is not free");
	if (m_currentTaskItemId != BAD_TASK_ITEM_ID)
	  logMsg(LOG_WARNING, "requested play toggle: task is empty but current task item id is not null");
	m_status = StatusFree;//Just to increase server stability;
	m_currentTaskItemId = BAD_TASK_ITEM_ID;
	logMsg(LOG_DEBUG, "The server is requested for play toggle, but there is no any task");
	return 0;
      }
    if (m_status == StatusFree)
      {
	if (m_currentTaskItemId != BAD_TASK_ITEM_ID)
	  logMsg(LOG_WARNING, "Requested play toggle operation: status is free, task is not empty but task item id is not null");
	logMsg(LOG_INFO, "Received play command, server is free but task is not empty, playing starting item");
	m_currentTaskItemId = m_currentTask.startingItemId();
	if (m_currentTaskItemId == BAD_TASK_ITEM_ID)
	  {
	    logMsg(LOG_WARNING, "Requested play toggle operation: status is free, task is not empty but cannot determine starting item id");
	    return 0;
  }
	m_lastPos = 0;
	m_totalSec = 0;
	m_status = StatusPlaying;
	for(ClientList::const_iterator it = m_clients.begin();it != m_clients.end();++it)
	  (*it)->notifyOnNewTaskItem(m_currentTaskItemId);
	m_consoleStatus.notifyOnNewTaskItem(m_currentTask, m_currentTaskItemId);
	goPlay();
	return 1;
      }
    if (m_status == StatusPlaying)
      {
	logMsg(LOG_INFO, "The server is requested to pause the playback");
	m_player->pause();
	m_status = StatusPaused;
	for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
	  (*it)->notifyOnIdle();
	m_consoleStatus.notifyOnIdle();
	return 0;
      }
    assert(m_status == StatusPaused);
    logMsg(LOG_INFO, "The server is requested to resume playback");
    m_status = StatusPlaying;
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
      (*it)->notifyOnNewPos(m_lastPos, m_totalSec);
    m_consoleStatus.notifyOnNewPos(m_lastPos, m_totalSec);
    m_player->resume();
    return 1;
  }

  void onStop()
  {
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested to stop playback, but there is no any task at all");
	return;
      }
    logMsg(LOG_INFO, "The server is requested to stop playback");
    m_player->stop();
    m_status = StatusFree;
    //Leaving current task content without changing;
    m_currentTaskItemId = BAD_TASK_ITEM_ID;
    m_lastPos = 0;
    m_totalSec = 0;
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
      (*it)->notifyOnTaskFinish();
    m_consoleStatus.notifyOnTaskFinish();
  }

  void onForward(size_t offset, size_t& newPos, size_t& total)
  {
    if (offset == 0)
      {
	logMsg(LOG_DEBUG, "The server is requested for forward jump with zero offset, doing nothing");
	newPos = m_lastPos;
	total = m_totalSec;
	return;
      }
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested for forward jump but there is no any task, doing nothing");
	newPos = 0;
	total = 0;
	return;
      }
    if (m_status == StatusPaused)
      {
	logMsg(LOG_DEBUG, "The server is requested for forward jump but playing process is paused, switching to playing mode first");
	m_player->resume();
	m_status = StatusPlaying;
      }
    assert(m_status == StatusPlaying);
    m_lastPos += offset;
    if (m_lastPos > m_totalSec)
      m_lastPos = m_totalSec;
    newPos = m_lastPos;
    total = m_totalSec;
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewPos(m_lastPos, m_totalSec);
    m_consoleStatus.notifyOnNewPos(m_lastPos, m_totalSec);
    logMsg(LOG_INFO, "The server is requested for forward jump on %u seconds", offset);
    m_player->jumpForward(offset);
  }

  void onBackward(size_t offset, size_t& newPos, size_t& total)
  {
    if (offset == 0)
      {
	logMsg(LOG_DEBUG, "The server is requested for backward jump with zero offset, doing nothing");
	newPos = m_lastPos;
	total = m_totalSec;
	return;
      }
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested for backward jump but there is no any task, doing nothing");
	newPos = 0;
	total = 0;
	return;
      }
    if (m_status == StatusPaused)
      {
	logMsg(LOG_DEBUG, "The server is requested for backward jump but playing process is paused, switching to playing mode first");
	m_player->resume();
	m_status = StatusPlaying;
      }
    assert(m_status == StatusPlaying);
    m_lastPos = m_lastPos > offset?m_lastPos - offset:0;
    total = m_totalSec;
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewPos(m_lastPos, m_totalSec);
    m_consoleStatus.notifyOnNewPos(m_lastPos, m_totalSec);
    logMsg(LOG_INFO, "The server is requested for backward jump on %u seconds", offset);
    m_player->jumpBackward(offset);
  }

  TaskItemId onNext()
  {
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested for next task item switching but there is no any task, doing nothing");
	return BAD_TASK_ITEM_ID;
      }
    const TaskItemId newTaskItemId = m_currentTask.nextTaskItemId(m_currentTaskItemId);
    if (newTaskItemId == BAD_TASK_ITEM_ID)
      {
	logMsg(LOG_DEBUG, "The server is requested for jump to next task item but there is no any");
	return BAD_TASK_ITEM_ID;
      }
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewTaskItem(newTaskItemId);
    m_consoleStatus.notifyOnNewTaskItem(m_currentTask, newTaskItemId);
    m_lastPos = 0;
    m_totalSec = 0;
    m_currentTaskItemId = newTaskItemId;
    logMsg(LOG_INFO, "The server is requested for next task item switching");
    goPlay();
    return newTaskItemId;
  }

  TaskItemId onPrevious()
  {
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested for previous task item switching but there is no any task, doing nothing");
	return BAD_TASK_ITEM_ID;
      }
    const TaskItemId newTaskItemId = m_currentTask.prevTaskItemId(m_currentTaskItemId);
    if (newTaskItemId == BAD_TASK_ITEM_ID)
      {
	logMsg(LOG_DEBUG, "The server is requested for jump to previous task item but there is no any");
	return BAD_TASK_ITEM_ID;
      }
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewTaskItem(newTaskItemId);
    m_consoleStatus.notifyOnNewTaskItem(m_currentTask, newTaskItemId);
    m_lastPos = 0;
    m_totalSec = 0;
    m_currentTaskItemId = newTaskItemId;
    logMsg(LOG_INFO, "The server is requested for previous task item switching");
    goPlay();
    return newTaskItemId;
  }

  bool onTaskItem(TaskItemId taskItemId)
  {
    assert(taskItemId != BAD_TASK_ITEM_ID);
    if (m_status == StatusFree)
      {
	logMsg(LOG_DEBUG, "The server is requested for switching to new task item but there is no any task, doing nothing");
	return 0;
      }
    if (!m_currentTask.hasTaskItem(taskItemId))
      {
	logMsg(LOG_DEBUG, "The server is requested for switching to non-existing task");
	return 0;
      }
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewTaskItem(taskItemId);
    m_consoleStatus.notifyOnNewTaskItem(m_currentTask, taskItemId);
    m_lastPos = 0;
    m_totalSec = 0;
    m_currentTaskItemId = taskItemId;
    logMsg(LOG_DEBUG, "The server is requested for switching to new task item");
    goPlay();
    return 1;
  }

  size_t onIncreaseVolume(size_t step)
  {
    m_currentVolume += step;
    if (m_currentVolume >= MAX_VOLUME_VALUE)
      m_currentVolume = MAX_VOLUME_VALUE;
    m_player->setVolume(m_currentVolume);
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewVolume(m_currentVolume);
    m_consoleStatus.notifyOnNewVolume(m_currentVolume);
    return m_currentVolume;
  }

  size_t onDecreaseVolume(size_t step)
  {
    m_currentVolume = step < m_currentVolume?m_currentVolume - step:0;
    m_player->setVolume(m_currentVolume);
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewVolume(m_currentVolume);
    m_consoleStatus.notifyOnNewVolume(m_currentVolume);
    return m_currentVolume;
  }

private: //AbstractSignalHandler;
  void onSystemSignal()
  {
    if (!wasSigChld)
      return;//there was not any signal, silently doing nothing;
    logMsg(LOG_DEBUG, "SIGCHLD registered, trying to collect all available zombies");
    pid_t pid = waitpid(-1, NULL, WNOHANG);
    while (pid > 0)
      {
	logMsg(LOG_DEBUG, "Collected zombie with pid=%d", pid);
	m_player->onSigChld(pid);
	pid = waitpid(-1, NULL, WNOHANG);
      }
  }

private://AbstractPlayerCallback;
  void onPlayerPosition(size_t secCurrent, size_t secTotal)
  {
    if (secCurrent == m_lastPos && secTotal == m_totalSec)
      return;
    m_lastPos = secCurrent;
    m_totalSec = secTotal;
    logMsg(LOG_DEBUG, "New player position: %u/%u", m_lastPos, m_totalSec);
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();++it)
      (*it)->notifyOnNewPos(m_lastPos, m_totalSec);
    m_consoleStatus.notifyOnNewPos(m_lastPos, m_totalSec);
  }

  void onPlayerStop(int exitCode)
  {
    if (m_status == StatusFree)
      {
	logMsg(LOG_WARNING, "Received notification the player has finished work but status marked as free. Doing nothing.");
	return;
      }
    if (m_status == StatusPaused)
      logMsg(LOG_WARNING, "The player has finished its work but status marked as paused");
    if (m_currentTaskItemId == BAD_TASK_ITEM_ID)
      {
	logMsg(LOG_ERR, "Player has finished its job but there is no any information about current task item, resetting state");
	m_status = StatusFree;
	m_lastPos = 0;
	m_totalSec = 0;
	return;
      }
    m_currentTaskItemId = m_currentTask.nextTaskItemId(m_currentTaskItemId);
    if (m_currentTaskItemId == BAD_TASK_ITEM_ID)
      {
	logMsg(LOG_INFO, "Task playback completed!");
	m_status = StatusFree;
	m_lastPos = 0;
	m_totalSec = 0;
	for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
	  (*it)->notifyOnTaskFinish();
	m_consoleStatus.notifyOnTaskFinish();
	return;
      }
    logMsg(LOG_INFO, "Switching to next task item");
    m_status = StatusPlaying;
    m_lastPos = 0;
    m_totalSec = 0;
    for(ClientList::iterator it = m_clients.begin();it != m_clients.end();it++)
      (*it)->notifyOnNewTaskItem(m_currentTaskItemId);
    m_consoleStatus.notifyOnNewTaskItem(m_currentTask, m_currentTaskItemId);
    goPlay();
  }

private: //Auxiliary methods;
  void goPlay()
  {
    assert(m_player.get() != NULL);
    m_player->stop();
    assert(m_currentTask.hasTaskItem(m_currentTaskItemId));
    m_player->play(m_currentTask.itemFileName(m_currentTaskItemId), m_currentVolume);
  }

  void closeSockets()
  {
    for(SocketList::iterator it = m_sockets.begin();it != m_sockets.end();++it)
      delete *it;
    m_sockets.clear();
  }

  void installSignalProcessing()
  {
    struct sigaction sa;
    sigaction(SIGCHLD, NULL, &sa);
    sa.sa_handler = sigChldHandler;
    sa.sa_flags |= SA_RESTART;
    sigaction(SIGCHLD, &sa, NULL);
    sigset_t blockedMask;
    sigemptyset(&blockedMask);
    sigaddset(&blockedMask, SIGCHLD);
    sigprocmask(SIG_BLOCK, &blockedMask, &m_origMask);
  }

private:
  const Configuration& m_configuration;
  auto_ptr<AbstractPlayer> m_player;
  char m_status;
  Task m_currentTask;
  TaskItemId m_currentTaskItemId;
  size_t m_currentVolume;
  size_t m_lastPos, m_totalSec;
  SocketList m_sockets;
  ClientList m_clients;
  ConsoleStatus m_consoleStatus;
  bool m_terminationFlag;
  sigset_t m_origMask;
}; //class MusitoriusServer;

void printLogo()
{
  std::cout << "Musitorius is the media server. Version: " << PACKAGE_VERSION << std::endl;
  std::cout << std::endl;
}

int run(const Configuration& configuration)
{
  if (!configuration.daemonMode)
    {
      initLogging(1, LOG_INFO);//FIXME:
      printLogo();
    } else 
    initLogging(0, LOG_DEBUG);//FIXME:
  if (!configuration.pidFile.empty())
    {
      std::ofstream os(configuration.pidFile.c_str());
      if (!os)
	logMsg(LOG_ERR, "Cannot save process pid to \'%s\'", configuration.pidFile.c_str()); else
	os << getpid() << std::endl;
    }
  try {
    MusitoriusServer musitoriusServer(configuration);
    musitoriusServer.run();
    return 0;
  }
  catch(const MusitoriusException& e)
    {
      if (!configuration.daemonMode)
	std::cerr << PREFIX << e.getMessage() << std::endl;
      logMsg(LOG_CRIT, "%s", e.getMessage().c_str());
      return 1;
    }
  catch(std::bad_alloc)
    {
      if (!configuration.daemonMode)
	std::cerr << PREFIX << "no enough free memory for server execution" << std::endl;
      logMsg(LOG_CRIT, "No enough free memory for server execution");
      return 1;
    }
  catch(...)
    {
      if (!configuration.daemonMode)
	std::cerr << PREFIX << "unexpected exception" << std::endl;
      logMsg(LOG_CRIT, "Unexpected exception");
      return 1;
    }
  assert(0);
  return 1;
}

void printHelp()
{
  printLogo();
  std::cout << "Usage:" << std::endl;
  std::cout << "\tmusitoriusd [OPTIONS]" << std::endl;
  std::cout << std::endl;
  std::cout << "List of valid options:" << std::endl;
  cmdLine.printHelp();
}

bool loadConfiguration(Configuration& c)
{
  if (cmdLine.used("daemon"))
    c.daemonMode = 1;
  if (cmdLine.used("socket"))
    {
      if (trim(cmdLine["socket"]).empty())
	{
	  std::cerr << PREFIX << "socket file name specified in the command line cannot be empty" << std::endl;
	  return 0;
	}
      c.socket = cmdLine["socket"];
    }
  if (cmdLine.used("pidfile"))
    {
      if (trim(cmdLine["pidfile"]).empty())
	{
	  std::cerr << PREFIX << "pid file name specified in the command line cannot be empty" << std::endl;
	  return 0;
	}
      c.pidFile = cmdLine["pidfile"];
    }
  return 1;
}

int main(int argc, char* argv[])
{
  if (!cmdLine.parse(argc, argv))
    return 1;
  if (cmdLine.used("help"))
    {
      printHelp();
      return 0;
    }
  Configuration conf;
  if (!loadConfiguration(conf))
    return 1;
  if (!conf.daemonMode)
    return run (conf);
  pid_t pid = fork();
  if (pid == (pid_t)-1)
    {
      std::cerr << PREFIX << "cannot create child process to run musitorius server in daemon mode:" << ERRNO_MSG << std::endl;
      return 1;
    }
  if (pid == (pid_t)0)
    exit(run(conf));
  return 0;
}
